
#include <avr/pgmspace.h>

#include <EEPROM.h>
#include <Wire.h>
#include <LiquidCrystal_PCF8574.h>
LiquidCrystal_PCF8574 lcd(0x27);

#define ADC_OUT 0           // channel for AC output sample
#define ADC_DCI 3           // DC current sense
#define ADC_HS  2           // Heat sink temp sensor
#define ADC_DCV 1           // DC supply volts sample
#define ADC_TOR 7           // toroid temp sensor
#define ADC_ACI 6           // AC current from SCR/current trans do-dar

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))



// these are produced from ADC counts multiplied by scale factors. See loop() , parse_it(), show_menu()  
float ac_setpoint, ac_output, dcv, dci, hs_temp,tor_temp, aci;

typedef struct  {
float dc_v_cal,
      dc_i_cal,
      ac_v_cal,
      ac_setpoint;
float fan_on_temp,fan_off_temp,over_temp,
      tor_fan_on_temp,tor_fan_off_temp,tor_over_temp,
      low_limit,lv_cutoff,
      overcurrent, lv_restart;
uint16_t freq;
uint8_t lv_enabled,
      oi_enabled,
      uv_enabled,
      ot_enabled,
      tor_ot_enabled,
      backlight_enabled;
float dc_i_cal_off,
      ac_i_cal,
      ac_i_cal_off,
      tor_ntc_off,
      hs_ntc_off,
      lv_time_below;
} _nvdata;
      
_nvdata nvd;
      
uint8_t lv_stopped,oen,last_oen, stop_reason;
uint16_t dbounce;
double kp, ki, kd;
float time_below;

char lcdbuf[82],*lbp,bufc;
String spstr = "";
uint8_t got_data;

float get_float(void);
int get_int(void);
float get_temp(int);
void init_vars(void);
void do_20chars(char *);

void get_biquad(float,int,int);
void biquad_filter(float *,int);

float aa[5][4],d1[5][4],d2[5][4];
float w0[5][4],w1[5][4],w2[5][4];
int bqnp[4];
//float filtered_adc[2];

void get_biquad(float fc, int np, int idx)
  {   // calculate biquad Butterworth L.P. filter coefficients. np = no. of poles, fc = cut off freq., idx = which filter to calc.
  float s,r;
  int i;  
  float a = tan(M_PI*fc);
  float a2 = a*a;
  bqnp[idx] = np;
  for(i=0; i<np; ++i)
    {
    r =  sin(3.141592926 * (2.0*i+1.0)/(4.0*(float)np));
    s = a2 + 2.0*a*r + 1.0;
    aa[i][idx] = a2/s;
    d1[i][idx] = 2.0*(1-a2)/s;
    d2[i][idx] = -(a2 - 2.0*a*r + 1.0)/s;
    w1[i][idx]=w2[i][idx]=w0[i][idx]=0.0;
    }
  }

void biquad_filter(float *val, int idx)
{ // do LP filter, taking input value pointer val, applying filter index by idx, store filtered result
  float s;
  int i;
  s = *val;
  for(i=0; i<bqnp[idx]; ++i)
    {
    w0[i][idx] = d1[i][idx]*w1[i][idx] + d2[i][idx]*w2[i][idx] + s;
    s = aa[i][idx]*(w0[i][idx] + 2.0*w1[i][idx] + w2[i][idx]);
    w2[i][idx] = w1[i][idx];
    w1[i][idx] = w0[i][idx];
    }
  *val = s;
}

void setup()
  {
  Serial.begin(115200);  

  get_biquad(0.01,4,0); // calc filter coeffs for AC output
  get_biquad(0.01,4,1); // calc filter coeffs for DC current
  get_biquad(0.01,4,2); // calc filter coeffs for AC current
  get_biquad(0.01,4,3); // calc filter coeffs for DC volts
  
  lcd.begin(20,4);      // LCD stuff
  lcd.clear();
  lcd.print("Nanoverter");
  delay(2000);
  lcd.clear();
  delay(1000);
  lbp = &lcdbuf[1];   // point local pointer to lcd buffer, used by strncpy(), sprintf()
  strncpy(lbp,    "---                 ",20);
  strncpy(lbp+40, "H.S.     C TOR     C",20);
  strncpy(lbp+20, "AC                  ",20);
  strncpy(lbp+60, "DC    A     V      W",20);
  bufc=0;

  pinMode(12,OUTPUT);     // heat sink fan
  pinMode(11,OUTPUT);     // toroid fan
  pinMode(8,INPUT);// momentary on/off button, normally open  
  pinMode(7,OUTPUT);
  pinMode(6,INPUT);       // let SCR_SD_N drive this, low impedence input 0 or 5V
  digitalWrite(6,HIGH);
  pinMode(4,OUTPUT);      // timing testing, 
  
  cbi(PORTD,7);           // on boot, pulled low for some reason, starting inverter. so pull high.
  lv_stopped = 2;         // prevent auto start on boot up.
  init_vars();
  sbi(ADCSRA,ADPS2) ;   // ADC clock prescale = 16 now. this saves 100us in main loop with little cost to loop stability
  cbi(ADCSRA,ADPS1) ;
  cbi(ADCSRA,ADPS0) ;
  }



void init_vars()      
  {
  dbounce = 0;    
  got_data = 0;
  get_nvdata();
  stop_reason = 0;
  // clear error if needed
  strncpy(lbp,"                    ",20);
  oen = 0;
  last_oen = 0;
  time_below = 0.0;
  }


float get_temp(int Vo) 
 {    // https://circuitdigest.com/microcontroller-projects/arduino-thermistor-interfacing-code-circuit
 float a = 1.009249522e-03, b = 2.378405444e-04, c = 2.019202697e-07;
 float T,logRt,Tc;
 //return Vo; 
 logRt = log(10000.0*((1024.0/Vo-1))); 
 T = (1.0 / (a + b*logRt + c*logRt*logRt*logRt)); // We get the temperature value in Kelvin from this Stein-Hart equation
 Tc = T - 273.15;                                 // Convert Kelvin to Celsius
 return Tc;
}


void loop()
{
  //
  // this runs at about 500Hz
  //
  float dcw,r;
  char fbuf[10];
  //
  // read inputs
  // oen = 1 if inverter output enabled, = 0 if stopped
  //
  sbi(PORTD,4);
  check_switch();   
  
  ac_output = (float)analogRead(ADC_OUT) * nvd.ac_v_cal; // get AC output voltage, scale it
  biquad_filter(&ac_output,0);    // then apply LP filter, to smooth out fast changing noise.
  
  dci = ((float)analogRead(ADC_DCI) - nvd.dc_i_cal_off) * nvd.dc_i_cal;
  biquad_filter(&dci,1);    // then apply LP filter, to smooth out fast changing noise.

  aci = ((float)analogRead(ADC_ACI) - nvd.ac_i_cal_off) *  nvd.ac_i_cal ;
  biquad_filter(&aci,2);    // then apply LP filter, to smooth out fast changing noise.
  
  dcv = (float)analogRead(ADC_DCV) * nvd.dc_v_cal;
  biquad_filter(&dcv,3); 
  
  hs_temp = get_temp(analogRead(ADC_HS)) - nvd.hs_ntc_off; 
  tor_temp = get_temp(analogRead(ADC_TOR)) - nvd.tor_ntc_off; 
  //
  // update LCD
  //
  if (oen == 1)
    strncpy(lbp,"ON ",3);
  else
    strncpy(lbp,"OFF",3);
  dcw = dci*dcv;  // calc watts

  strcpy(lbp+22,"         ");
  dtostrf(aci,2,1,fbuf);
  strcpy(lbp+27 - strlen(fbuf),fbuf);
  *(lbp+27)='A';
  
  dtostrf(ac_output,2,1,fbuf);
  strcpy(lbp+34 - strlen(fbuf),fbuf);
  *(lbp+34)='V';

 
  sprintf(lbp+45,"%3d",(int)hs_temp);*(lbp+48)=' ';*(lbp+49)='C';
  sprintf(lbp+55,"%3d",(int)tor_temp);*(lbp+58)=' ';*(lbp+59)='C';

  strcpy(lbp+62,"     ");
  dtostrf(dci,2,1,fbuf);
  strcpy(lbp+67 - strlen(fbuf),fbuf);
  *(lbp+67)='A';
  
  dtostrf(dcv,2,1,fbuf);
  strcpy(lbp+73 - strlen(fbuf),fbuf);
  *(lbp+73)='V';
  
  sprintf(lbp+74,"%5d",(int)dcw);*(lbp+79)='W';

  //
  // do limit checking. Probably need some timeouts too. The inputs will be noisy.
  //
  if (nvd.oi_enabled == 1 && dci > nvd.overcurrent)    // if overcurrent limit enabled check if over current
    {
    oen = 0;
    strncpy(lbp+5,"DC O.C. SD   ",12);
    stop_reason = 1;
    }
  if (lv_stopped == 0 && nvd.lv_enabled == 1 && dcv < nvd.lv_cutoff)
    {
      time_below += 0.002;  // 1/500th second
      if (time_below > nvd.lv_time_below)
        {
        oen = 0;
        strncpy(lbp+5,"DC U.V. stop",12);
        lv_stopped = 1;
        stop_reason = 2;
        }
    }
  if (lv_stopped == 0 && nvd.lv_enabled == 1 && dcv > nvd.lv_cutoff) 
    time_below = 0.0;       // reset timer

  // remove this, debugging
  //strcpy(lbp+6,"         ");
  //dtostrf(time_below,2,2,fbuf);
  //strcpy(lbp+11 - strlen(fbuf),fbuf);
  //*(lbp+12)=' ';
  //

     
  if (lv_stopped == 1 && (dcv > nvd.lv_restart) && oen == 0 )
    {
    lv_stopped = 0;
    init_vars();
    oen = 1;    // enable it.
    }
  if (nvd.ot_enabled == 1 && hs_temp > nvd.over_temp)
    {
    oen = 0;
    strncpy(lbp+5,"H.S. OT SD",10);
    stop_reason = 3;
    }
  if (nvd.tor_ot_enabled == 1 && tor_temp > nvd.tor_over_temp)
    {
    oen = 0;
    strncpy(lbp+5,"Tor OT SD ",10);
    stop_reason = 6;
    }
    //
    // need a delay timers after eon = 1 for the following, else it will always be under voltage
    //
  if (nvd.uv_enabled == 1 && (ac_output < nvd.low_limit))
    {
    oen = 0;
    strncpy(lbp+5,"AC U.V. SD    ",14);  
    stop_reason = 4;
    }
  
  if ((PIND & 0x40) == 0x40)      // D6 goes high
    {
    oen = 0;
    strncpy(lbp+5,"SCR OC SD     ",14);  
    stop_reason = 5;
    } 
  
  if (hs_temp > nvd.fan_on_temp)
    {
    lbp[44] = '*';
    sbi(PORTB,4);  
    }
  if ( hs_temp < nvd.fan_off_temp)
    {
    lbp[44] = ' ';
    cbi(PORTB,4);  
    }
  if (tor_temp > nvd.tor_fan_on_temp)
    {
    lbp[54] = '*';
    sbi(PORTB,3);  
    }
  if (tor_temp < nvd.tor_fan_off_temp)
    {
    lbp[54] = ' ';
    cbi(PORTB,3);  
    }  
  //
  // drive nano1 on/off control
  //
  if (oen == 1)
    sbi(PORTD,7);
  else
    cbi(PORTD,7);
   //
   // update LCD
   // 
  bufc++;
  poke(lcdbuf[bufc],1);
  if (bufc > 79)
        {
        lcdxy(0,0);
        bufc = 0;
        }
  // this will pause loop() processing, i.e. it is blocking
  if (got_data == 1)
    parse_it();
  if (oen == 1)
    stop_reason = 0;  
  cbi(PORTD,4);
}

void check_switch()
{
  dbounce++;  
  if (dbounce > 250) dbounce = 250;
  if ((PINB & 1) == 0)     // momentary button pressed, pin D8, pulling it low
    {
    if (dbounce > 200)
      { 
      lv_stopped = 0;   
      if(oen == 0)
        {
        init_vars();
        oen = 1;    // enable it.
        }
      else
        oen = 0;   // stop inverter
      dbounce=0;
      }       
    }
}




/*
 *  config menu and serial port stuff. Only active when oen = 0
 */
void show_menu()
  {
    
  char *p;
  get_nvdata();
  spstr = "";
  got_data=0;
  Serial.println(F("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"));
  Serial.print(F("heat sink temp = "));Serial.print(hs_temp,0);Serial.print(" degC    ch.");Serial.println(ADC_HS);
  Serial.print(F("toroid temp = "));Serial.print(tor_temp,0);Serial.print(" degC    ch.");Serial.println(ADC_TOR);
  Serial.print(F("1 - cal.  DC volts      ")); Serial.print(nvd.dc_v_cal);Serial.print(" -> ");Serial.print(dcv,2);Serial.print(" V    ch.");Serial.println(ADC_DCV);
  Serial.print(F("2 - cal.  AC current    ")); Serial.print(nvd.ac_i_cal);Serial.print(" -> ");Serial.print(aci,2);Serial.print(" V    ch.");Serial.println(ADC_ACI);
  Serial.print(F("3 - cal. ACI offset     ")); Serial.print(nvd.ac_i_cal_off);Serial.println(" ADC counts");
  Serial.print(F("4 - cal.  DC current    ")); Serial.print(nvd.dc_i_cal);Serial.print(" -> ");Serial.print(dci,1);Serial.print(" A    ch.");Serial.println(ADC_DCI);
  Serial.print(F("5 - cal. DCI offset     ")); Serial.print(nvd.dc_i_cal_off);Serial.println(" ADC counts");
  Serial.print(F("6 - cal.  AC output     ")); Serial.print(nvd.ac_v_cal);Serial.print(" -> ");Serial.print(ac_output,0);Serial.print(" V AC   ch.");Serial.println(ADC_OUT);
  Serial.print(F("N - Toroid NTC offset   ")); Serial.print(nvd.tor_ntc_off,0);Serial.println(" degC");
  Serial.print(F("7 - Toroid fan ON       ")); Serial.print(nvd.tor_fan_on_temp,0);Serial.println(" degC");
  Serial.print(F("8 - Toroid fan OFF      ")); Serial.print(nvd.tor_fan_off_temp,0);Serial.println(" degC");
  Serial.print(F("9 - Toroid shutdown     ")); Serial.print(nvd.tor_over_temp,0);Serial.println(" degC");
  Serial.print(F("P - H.S. NTC offset     ")); Serial.print(nvd.hs_ntc_off,0);Serial.println(" degC");
  Serial.print(F("B - H.S. fan ON         ")); Serial.print(nvd.fan_on_temp,0);Serial.println(" degC");
  Serial.print(F("C - H.S  fan OFF        ")); Serial.print(nvd.fan_off_temp,0);Serial.println(" degC");
  Serial.print(F("D - H.S. shutdown       ")); Serial.print(nvd.over_temp,0);Serial.println(" degC");
  Serial.print(F("E - AC output low limit ")); Serial.print(nvd.low_limit,0);Serial.println(" V");
  Serial.print(F("F - LV cutoff voltage   ")); Serial.print(nvd.lv_cutoff,1);Serial.println(" V");
  Serial.print(F("Q - LV time below       ")); Serial.print(nvd.lv_time_below,1);Serial.println(" seconds");
  Serial.print(F("A - Restart voltage     ")); Serial.print(nvd.lv_restart,1);Serial.println(" V");
  Serial.print(F("G - over current limit  ")); Serial.print(nvd.overcurrent,1);Serial.println(" A");
  Serial.print(F("H - enable/disable DC LV cutoff             "));  if (nvd.lv_enabled == 1)  Serial.println("ON"); else Serial.println("OFF");  
  Serial.print(F("I - enable/disable over current cutoff      "));  if (nvd.oi_enabled == 1)  Serial.println("ON"); else Serial.println("OFF");  
  Serial.print(F("J - enable/disable AC under volt cutoff     "));  if (nvd.uv_enabled == 1)  Serial.println("ON"); else Serial.println("OFF");  
  Serial.print(F("K - enable/disable H.S. over temp cutoff    "));  if (nvd.ot_enabled == 1)  Serial.println("ON"); else Serial.println("OFF");  
  Serial.print(F("L - enable/disable Toroid over temp cutoff  "));  if (nvd.tor_ot_enabled == 1)  Serial.println("ON"); else Serial.println("OFF");  
  Serial.print(F("M - enable/disable LCD backlight            "));  if (nvd.backlight_enabled == 1)  Serial.println("ON"); else Serial.println("OFF");  
  Serial.print(F("Z - zero E2PROM")); Serial.println();
  Serial.print(F("Y - put defaults into E2PROM")); Serial.println();
  Serial.println();
  }


void serialEvent() 
{
  char c;
  while (Serial.available()) 
    {
    c = (char)Serial.read();
    if (got_data == 0)    // dont store while still working on previous string
      {
      spstr += c;
      if (c == '\n')
        got_data = 1;
      }
    }
}

void do_20chars(char *p)
  {
    int i;
    for(i=0; i < 20; i++)
      Serial.print(p[i]);
    Serial.println();
  }

int adc_average(int ch, int n)
  {
  float tv;
  int i;
  tv = 0.0;
  for(i=0; i < n; i++)
    tv += (float)analogRead(ch);
  tv /= (float)n;
  return (int)tv;
  }

void parse_it()
  {
  int i;
  float f,v;
  char c,*p;
  spstr.toUpperCase();
  c = spstr.charAt(0);
  switch (c)
    { 
    case '^':         // this probably used by myself only, to allow remote control of nanverter via the raspberry pi & USB serial
      spstr.setCharAt(0,' ');                           // remove ^, so as to allow conversion to int
      i = spstr.toInt();                                // send ^nn<newline>
      if (i == 42)        Serial.println("the answer to life, the universe and everything");
      if (i == 10)        oen = 0;                      // 
      if (i == 11)        oen = 1;                      // eg 13<newline> to obtain DC volts
      if (i == 12)        Serial.println(ac_output);    // or 11<newline> to start inverter, etc.
      if (i == 13)        Serial.println(dcv);
      if (i == 14)        Serial.println(dci);
      if (i == 15)        Serial.println(hs_temp);
      if (i == 16)        Serial.println(nvd.overcurrent);
      if (i == 17)        Serial.println(nvd.over_temp);
      if (i == 18)        Serial.println(nvd.lv_cutoff);
      if (i == 19)        Serial.println(nvd.lv_restart);
      if (i == 20)        Serial.println(nvd.oi_enabled);
      if (i == 21)        Serial.println(nvd.ot_enabled);
      if (i == 22)        Serial.println(nvd.lv_enabled);
      if (i == 23)        Serial.println(tor_temp);
      if (i == 24)        Serial.println(nvd.tor_ot_enabled);
      if (i == 30)        Serial.println(oen);
      if (i == 31)        Serial.println(stop_reason);   
      if (i == 32)        Serial.println((PINB));         // D11,D12 is fan control. look for bit 3,4
      if (i == 33)        // dump LCD buffer
        {
          do_20chars(lbp);
          do_20chars(lbp + 40);
          do_20chars(lbp + 20);
          do_20chars(lbp + 60);
        }
      break;      
    case '?':
      show_menu();
      break;
    case 'Z':      // clear e2prom
      Serial.println("clearing e2prom..");
      for ( i = 0 ; i < EEPROM.length() ; i++) 
        EEPROM.write(i, 0);
      Serial.println("done");
      nv();
      break;
    case 'Y':
      do_nvdefaults();
      nv();
      break;      
    case '1': // cal DC voltage
      Serial.println(F("DC cal. Enter volts"));
      f = get_float();
      i = analogRead(ADC_DCV);
      nvd.dc_v_cal = f / (float)i;
      nv();
      break;   
    case '2': // cal AC current
      Serial.println(F("AC cal. enter Amps "));
      f = get_float();
      //i = analogRead(ADC_ACI);
      i = adc_average(ADC_ACI,25000);
      
      nvd.ac_i_cal = f / ((float)i - nvd.ac_i_cal_off);
      nv();
      break;  
    case '3': // cal AC current offset
      Serial.println(F("ACI offset, apply zero amps "));
      i = analogRead(ADC_ACI);
      nvd.ac_i_cal_off = (float)i;
      nv();
      break;         
    case '4': // cal DC current
      Serial.println(F("DC current cal. Enter amps"));
      f = get_float();
      //i = analogRead(ADC_DCI);
      i = adc_average(ADC_DCI,25000);
      
      nvd.dc_i_cal = f/ ((float)i - nvd.dc_i_cal_off);
      nv();
      break;      
    case '5': // cal DC current sense offset
      Serial.println(F("DCI offset, apply zero amps "));
      i = analogRead(ADC_DCI);
      nvd.dc_i_cal_off =  (float)i;
      nv();
      break;       
    case '6': 
      Serial.println(F("AC output cal. Enter volts"));
      f = get_float();
      //i = analogRead(ADC_OUT);
      i = adc_average(ADC_OUT,25000);
      
      nvd.ac_v_cal = f/(float)i;
      nv();
      break;    
    case 'N': //  toriid NCT offset
      Serial.println(F("Toroid NTC offset (degC)"));
      nvd.tor_ntc_off = get_float();
      nv();
      break;       
    case '7': //  toroid fan ON temp
      Serial.println(F("Toroid Fan ON temperature (degC)"));
      nvd.tor_fan_on_temp = get_float();
      nv();
      break;    
    case '8': //  toroid fan OFF temp
      Serial.println(F("Toroid Fan OFF temperature (degC)"));
      nvd.tor_fan_off_temp = get_float();
      nv();
      break;   
    case '9': //  cut off temp
      Serial.println(F("Toroid Shut down temperature (degC)"));
      nvd.tor_over_temp = get_float();
      nv();
      break;                   
    case 'A': //  restart volts
      Serial.println(F("Restart after LV stop. Enter Volts"));
      nvd.lv_restart = get_float();
      nv();
      break; 
    case 'Q': //  time that DC needs to be less than DC cut off
      Serial.println(F("LV time DC has to remain below limit. Enter Seconds"));
      nvd.lv_time_below = get_float();
      nv();
      break;   
    case 'P': //  h.s. offset
      Serial.println(F("H.S. NTC offset (degC)"));
      nvd.hs_ntc_off = get_float();
      nv();
      break;      
    case 'B': //  fan ON temp
      Serial.println(F("H.S. Fan ON temperature (degC)"));
      nvd.fan_on_temp = get_float();
      nv();
      break;    
    case 'C': //  fan OFF temp
      Serial.println(F("H.S. Fan OFF temperature (degC)"));
      nvd.fan_off_temp = get_float();
      nv();
      break;   
    case 'D': //  cut off temp
      Serial.println(F("H.S. Shut down temperature (degC)"));
      nvd.over_temp = get_float();
      nv();
      break;     
    case 'E': //  AC output low
      Serial.println(F("AC output Low Limit. Enter Volts"));
      nvd.low_limit = get_float();
      nv();
      break; 
    case 'F': //  DC input low
      Serial.println(F("LV shut down voltage. Enter volts"));
      nvd.lv_cutoff = get_float();
      nv();
      break; 
    case 'G': //  over current
      Serial.println(F("Over current limit. Enter Amps"));
      nvd.overcurrent = get_float();
      nv();
      break;
    case 'H': //  enable DC LV
      Serial.println(F("DC LV cutoff (1 = enable,0 = disable)"));
      nvd.lv_enabled = (get_int() > 0);
      nv();
      break; 
    case 'I': //  enable overcurrent
      Serial.println(F("Overcurrent cutoff (1 = enable,0 = disable)"));
      nvd.oi_enabled = (get_int() > 0);
      nv();
      break; 
    case 'J': //  enable AC UV
      Serial.println(F("AC under volt cutoff (1 = enable,0 = disable)"));
      nvd.uv_enabled = (get_int() > 0);
      nv();
      break;     
    case 'K': //  enable HS over temp
      Serial.println(F("H.S. over temp cutoff (1 = enable,0 = disable)"));
      nvd.ot_enabled = (get_int() > 0);
      nv();
      break;    
    case 'L': //  enable toroid over temp
      Serial.println(F("Toroid over temp cutoff (1 = enable,0 = disable)"));
      nvd.tor_ot_enabled = (get_int() > 0);
      nv();
      break;  
    case 'M': //  enable LCD backlight
      Serial.println(F("LCD backlight (1 = enable,0 = disable)"));
      nvd.backlight_enabled = (get_int() > 0);
      nv();
      break;                                                               
    default:   
      spstr="";
      got_data = 0;
  }
  spstr="";
  got_data = 0;
 }

void nv()
  {
    EEPROM.put(10,nvd);
    EEPROM.get(10,nvd);
    get_nvdata();
    show_menu();
  }

float get_float()
  {
    char c;
    got_data = 0;
    spstr = "";
    while (got_data == 0)
      {
      if (Serial.available()) 
        {
        c = (char)Serial.read();
        spstr += c;
        if (c == '\n')
          got_data = 1;
        }
      }
  return spstr.toFloat();
  }

int get_int()
  {
    char c;
    got_data = 0;
    spstr = "";
    while (got_data == 0)
      {
      if (Serial.available()) 
        {
        c = (char)Serial.read();
        spstr += c;
        if (c == '\n')
          got_data = 1;
        }
      }
  return spstr.toInt();
  }

void get_nvdata()
{
 EEPROM.get(10,nvd);
}

void do_nvdefaults()
{
 EEPROM.put(10,nvd);  // does nothing. Maybe it's best we are forced to set things as we need them?
}


void lcdxy( uint8_t col, uint8_t row)
  {
  int row_offsets[] = { 0x00, 0x40, 0x14, 0x54   }; 
  poke(0x80 | (col + row_offsets[row]),0);
  }

void poke(uint8_t c,uint8_t cmd)
  {
  uint8_t v2 = c    & 0x0F;
  uint8_t v1 = c >> 4 & 0x0F;
  uint8_t b,bl;
  bl = 0;
  if (nvd.backlight_enabled)
    bl = 0x08;
  Wire.beginTransmission(0x27);
  b = (v1 << 4) + 4 + cmd + bl;     // the + 0x08 will turn on LCD backlight
  Wire.write(b);
  _delay_us(2);
  b = (v1 << 4) + cmd;
  Wire.write(b);
  _delay_us(37);
  b = (v2 << 4) + 4 + cmd + bl;
  Wire.write(b);
  _delay_us(2);
  b = (v2 << 4) + cmd;
  Wire.write(b);
   _delay_us(37);
  Wire.endTransmission(); 
  }
